#!/usr/bin/perl
# -----------------------------------------------------------------------------
#  ident "@(#)parser.pl 1.2 04/02/06 SMI"
# -----------------------------------------------------------------------------
#
#  Copyright 2001-2004 Sun Microsystems, Inc. All rights reserved.
#
# -----------------------------------------------------------------------------

package PARSER;

#use strict;
require Text::ParseWords;
use variables;

use Exporter;
@Exporter::ISA = qw(Exporter);
@EXPORT = (CMD_SIMPLE);

require "$ENV{NHINSTALL_LIB}/tools/lib/common.pl";

#-----------------------------------------------------------------------------
#
# exported variables
#
#-----------------------------------------------------------------------------

# category of commands
$CMD_SIMPLE  = 0;
$CMD_FIELD   = 1;
$CMD_ARRAY   = 2;
$CMD_HASH    = 3;

# type of arguments
$CMD_BOOLEAN      = 0; # yes, no, true, false
$CMD_INTEGER      = 1; # integer or in a range
$CMD_DOTTED_IP    = 2; # dotted syntax
$CMD_NIC          = 3; # physical interface or logical interface
$CMD_CLIENT_ID    = 4; # MAC Address or client id (string)
$CMD_STRING       = 5; # free
$CMD_HOST         = 6; # limited set of characters
$CMD_LIST         = 7; # list of value separated by a comma
$CMD_BOOLEAN_YN   = 8; # only Y or N
$CMD_VENDOR       = 9; # is a vendor string

# for CMD_HOST
$CMD_PARTIAL_HOST     = 0; # is a postfix
$CMD_FULL_HOST        = 1; # is a host name

#-----------------------------------------------------------------------------
#
# general variables
#
#-----------------------------------------------------------------------------

my @G_command;
my $G_directive;
my $G_lineno;
my $G_filename;
my $G_parsing_ok = $TRUE;
my $G_cmd_error = $FALSE;

# command mode
my $PER_FIELD = 0;
my $PER_KEYWORD = 1;
my $G_mode = $PER_FIELD;

# special characters
my $optional_char = "-";
my $delimiter_string = "=";
my $continuation_string = "\\";

#==============================================================================
#
#   ERROR PROCESSING
#
#==============================================================================

#------------------------------------------------------------------------------
#
# print_error_no
#
#------------------------------------------------------------------------------

sub print_error_no {
  my ($message) = @_;
  
  $G_parsing_ok = $FALSE;
  $G_cmd_error = $TRUE;
 
  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime;
  my $date = sprintf("%02.02d:%02.02d:%02.02d", $hour, $min, $sec) ;
  COMMON::print_err(
	       "$date Error: $message (line $G_lineno, file $G_filename)\n") ;
}

#==============================================================================
#
#   COMMAND PARSER
#
#==============================================================================

#------------------------------------------------------------------------------
#
# extract a string from quotes
#
# return TRUE if it was a quoted string
#
#------------------------------------------------------------------------------

sub extract_string {

  my ($refvalue) = @_;

  my @array;

  if ($$refvalue =~ /^"(.*)"$/) { 
    @array = Text::ParseWords::parse_line('\s+', 0, $$refvalue);
    $$refvalue = $array[0];
    return $TRUE;
  }

  return $FALSE;
}

#------------------------------------------------------------------------------
#
# extract a string from quotes and keep special characters
#
# return TRUE if it was a quoted string
#
#------------------------------------------------------------------------------

sub extract_string_keep {

  my ($refvalue) = @_;

  my @array;

  if ($$refvalue =~ /^"(.*)"$/) { 
    @array = Text::ParseWords::parse_line('\s+', 1, $$refvalue);
    $$refvalue = $array[0];
    # remove external quote
    $$refvalue=substr($$refvalue, 1, length($$refvalue)-2);
    return $TRUE;
  }

  return $FALSE;
}

#------------------------------------------------------------------------------
#
# check a string
#
#------------------------------------------------------------------------------

sub is_valid_string {

  my ($refvalue, $refreason) = @_;

  my $is_a_string = extract_string($refvalue);

  if (length($$refvalue) == 0) { 
    $$refreason = "$$refvalue cannot be an empty string";
    return $FALSE;
  }

  return $TRUE;
}

#------------------------------------------------------------------------------
#
# check a boolean value and return a generic boolean
#
#------------------------------------------------------------------------------

sub is_boolean {

  my ($refvalue, $type, $refreason) = @_;

  my $upvalue = $$refvalue;
  $upvalue =~ tr/a-z/A-Z/;

  if ($type == $CMD_BOOLEAN) {

    if (($upvalue eq "Y") 
	or ($upvalue eq "T")
	or ($upvalue eq "YES")
	or ($upvalue eq "TRUE")) {
      $$refvalue = $TRUE;
      return $TRUE;
    } elsif  (($upvalue eq "N") 
	      or ($upvalue eq "F") 
	      or ($upvalue eq "NO") 
	      or ($upvalue eq "FALSE")) {
      $$refvalue = $FALSE;
      return $TRUE;
    }

    $$refreason = "$$refvalue not a boolean";

  } else {

    if ($upvalue eq "Y") {
      $$refvalue = $TRUE;
      return $TRUE;
    } elsif  ($upvalue eq "N") {
      $$refvalue = $FALSE;
      return $TRUE;
    }

    $$refreason = "$$refvalue (only Y or N allowed)";
  }

  return $FALSE;
}

#------------------------------------------------------------------------------
#
# check an integer (in a range if the control is present)
#
#------------------------------------------------------------------------------

sub is_integer {

  my ($refvalue, $control, $refreason) = @_;

  my $min;
  my $max;

  # check if it is an integer
  if ($$refvalue !~ /^\d+$/) {
    $$refreason = "$$refvalue not an integer";
    return $FALSE;
  }
  

  #check if there is in a range
  if ($control ne "") {

    ($min, $max) = split(/\.+/, $control);

    if ($min eq "") {
      
      # control= ..<max>
      if ($$refvalue > $max) {
	$$refreason = "$$refvalue incorrect (maximum value: $max)";
	return $FALSE;
      }
      
    } elsif ($max eq "") {
      
      # control= <min>..
      if ($$refvalue < $min) {
	$$refreason = "$$refvalue incorrect (minimum value: $min)";
	return $FALSE;
      }
      
    } else {
      
      # control = <min>..<max>
      if (($$refvalue < $min) || ($$refvalue > $max)) {
	$$refreason = "$$refvalue not in the range $min..$max";
	return $FALSE;
      }
    }
  }

  return $TRUE;
}

#------------------------------------------------------------------------------
#
# check a network interface and remove extra ":" if present
#
#------------------------------------------------------------------------------

sub is_nic {

  my ($refvalue, $refreason) = @_;

  # check something like hme0, hme0:, hme0:10
  if (($$refvalue !~ /^[a-zA-Z]+[0-9]+[:]?$/) &&
      ($$refvalue !~ /^[a-zA-Z]+[0-9]+:[0-9]*$/)) {
    $$refreason = "$$refvalue does not look like a network interface";
    return $FALSE;
  }

  if ($$refvalue =~ /^([a-zA-Z]+[0-9]+):$/) {
    # remove ":"
    $$refvalue = $1;
  }
  
  return $TRUE;
}

#------------------------------------------------------------------------------
#
# check if it's a MAC address or a client ID
#
#------------------------------------------------------------------------------

sub is_client_id {

  my ($refvalue, $refreason) = @_;

  my $value;
  my $nb = 0;

  my $savvalue = $$refvalue;
  my $is_a_string = extract_string($refvalue); 

  # check something like xx:xx:xx:xx:xx:xx
  if ((! $is_a_string) && ($$refvalue =~ /^([^:]*):/)) {
    foreach $value (split(":", $$refvalue)) {
      $nb++;
      if (! COMMON::is_hexa(\$value, 2, $refreason)) {
	return $FALSE;
      }
    }
    if ($nb != 6) {
      $$refreason = "$$refvalue: invalid number of hexadecimal values (exactly 6 are required)";
      return $FALSE;
    }
    # convert uppercase into lowercase (required for Jumpstart)
    $$refvalue =~ tr/A-Z/a-z/;
  } else {
    if ( ! is_valid_string($refvalue, $refreason)) {
      return $FALSE ;
    }

    # reuse the input value and process it to keep special characters
    $$refvalue = $savvalue;
    extract_string_keep($refvalue);
    return COMMON::string_hexa($$refvalue, \$hexa, $refreason);
  }

  return $TRUE;
}


#------------------------------------------------------------------------------
#
# check a host name
#
# partial mean that it's a postfix: starting with a alphabetic character
# is not required
#
#------------------------------------------------------------------------------

sub is_hostname {

  my ($refvalue, $control, $refreason) = @_;

  my $is_a_string = extract_string($refvalue); 

  if ($control == $CMD_FULL_HOST) {
    # cannot be empty
    if (length($$refvalue) == 0) {
      $$refreason = "cannot be an empty string";
      return $FALSE;
    }
    # check if it is a host name
    if ($$refvalue !~ /^[a-zA-Z]+[a-zA-Z.0-9-]*$/) {
      $$refreason = "$$refvalue does not look like a host name";
      return $FALSE;
    }
  } else {
    # check if it is a postfix (can be empty)
    if ($$refvalue !~ /^[a-zA-Z.0-9-]*$/) {
      $$refreason = "$$refvalue does not look like a postfix name";
      return $FALSE;
    }
  }
  
  return $TRUE;
}

#------------------------------------------------------------------------------
#
# check a vendor string
#
# no space nor comma
#
#------------------------------------------------------------------------------

sub is_vendor {

  my ($refvalue, $control, $refreason) = @_;

  my $is_a_string = extract_string($refvalue); 

  if (length($$refvalue) <= 0) {
    $$refreason = "cannot be an empty string";
    return $FALSE;
  }

  # look for a comma - space
  if ($$refvalue =~ m/[,\s\t]/) {
    $$refreason = "not a valid vendor string";
    return $FALSE;
  }
    
  
  return $TRUE;
}

#------------------------------------------------------------------------------
#
# check an IP address
#
#------------------------------------------------------------------------------

sub is_ip {

  my ($refvalue, $refreason) = @_;

  # check if it is a valid IP address
  if ($$refvalue !~ /^([0-9]+).([0-9]+).([0-9]+).([0-9]+)$/) {
    $$refreason = "$$refvalue does not look like an IP address";
    return $FALSE;
  }
  if (($1 < 0) || ($1 > 255)
      || ($2 < 0) || ($2 > 255)
      || ($3 < 0) || ($3 > 255)
      || ($2 < 0) || ($4 > 255)) {
    $$refreason = "$$refvalue does not look like an IP address";
    return $FALSE;
  }
  
  return $TRUE;
}

#------------------------------------------------------------------------------
#
# check if the string is in a list and return the uppercase value
#
#------------------------------------------------------------------------------

sub is_in_list {

  my ($refvalue, $control, $refreason) = @_;

  my $value;

  my $is_a_string = extract_string($refvalue);

  $$refvalue =~ tr/a-z/A-Z/;

  # check if the value is in the list
  foreach $value (split(",",$control)) {
    $value =~ tr/a-z/A-Z/;
    if ($$refvalue eq $value) {
      return $TRUE;
    }
  }
  
  $$refreason = "$$refvalue not in $control";
  return $FALSE;
}

#------------------------------------------------------------------------------
#
# build_label
#
#  return the information to be displayed regarding the directive
#  the field and so on
#
#------------------------------------------------------------------------------

sub build_label {

  my ($cmdidx, $argidx, $arguments, $nb_field) = @_;

  my $label;
  my $meaning;
  my $category = $G_command[$cmdidx]{category};

  # prepare the information to be displayed in case of error
  if ($category == $CMD_HASH) {
    $meaning = $G_command[$cmdidx]{argument}[$argidx]{meaning};
    # first field is the key
    if ($nb_field > 0) {
      $label = "$G_directive @$arguments[0]: $meaning";
    } else {
      $label = "$G_directive: $meaning";
    }
  } elsif ($category != $CMD_SIMPLE) {
    $meaning = $G_command[$cmdidx]{argument}[$argidx]{meaning};
    $label = "$G_directive: $meaning";
  } else {
    $label = "$G_directive";
  }

  return $label;
  
}

#------------------------------------------------------------------------------
#
# analyze the command per field
#
#------------------------------------------------------------------------------

sub analyze_per_field {

  my ($cmdidx, $arguments) = @_;

  my $argidx;
  my $arg;
  my $for_array;
  my $value;
  my $field;
  my $ref;

  my $nb_field = @{$arguments};
  my $nb_arg = @{$G_command[$cmdidx]{argument}};
  my $category = $G_command[$cmdidx]{category};
  my $reason;
  my $label;
  my $type;
  my $control;

  # reset the indicator of error in the command
  $G_cmd_error = $FALSE;

  # initialize all the impacted variables or fields 
  # (use default value if it exists)
  for ($argidx = 0 ; $argidx < $nb_arg; $argidx++) {
    $value = "";
    if (defined($G_command[$cmdidx]{argument}[$argidx]{default})) {
      $value = $G_command[$cmdidx]{argument}[$argidx]{default};
    }
    if (($category == $CMD_SIMPLE) || ($category == $CMD_FIELD)) {
      
      $ref = $G_command[$cmdidx]{argument}[$argidx]{variable};
      $$ref = $value;
    } elsif ($category == $CMD_HASH) {
      # it's an hash
      $field=$G_command[$cmdidx]{argument}[$argidx]{field};
      $for_array->{$G_command[$cmdidx]{argument}[$argidx]{field}} = $value;
    }
  }

  #
  # process all the field present in the command
  #
  for ($argidx = 0; ($argidx < $nb_field) && ($argidx < $nb_arg); $argidx++) {	
      
    $type    = $G_command[$cmdidx]{argument}[$argidx]{type};
    $control = $G_command[$cmdidx]{argument}[$argidx]{control};

    $label = build_label($cmdidx, $argidx, $arguments, $nb_field);

    # the value can be modified by the various control functions
    $value   = @$arguments[$argidx];

    # check if the arg is mandatory and has been replaced with optional char
    # if this is an authorized optional, replace it with a null string

    if ($value eq $optional_char) {
      if ($G_command[$cmdidx]{argument}[$argidx]{mandatory}) {
	print_error_no("$label is mandatory") ;
	next;
      } else {
	$value = "";
      }
    } else {

      # check if the argument is correct
      # the argument can be replaced by another value (for example boolean)
      # or convert (string into an integer)

      
      
      if (($type eq $CMD_BOOLEAN) || ($type == $CMD_BOOLEAN_YN)) {
	if (! is_boolean(\$value, $type, \$reason)) {
	  print_error_no("$label: $reason");
	  next ;
	}
      } elsif ($type eq $CMD_STRING) {
	if (! is_valid_string(\$value, \$reason)) {
	  print_error_no("$label: $reason");
	  next ;
	}
      } elsif ($type eq $CMD_INTEGER) {
	if (! is_integer(\$value, $control, \$reason)) {
	  print_error_no("$label: $reason");
	  next ;
	} 
      } elsif ($type eq $CMD_NIC) {
	if (! is_nic(\$value, \$reason)) {
	  print_error_no("$label: $reason");
	  next ;
	}
      } elsif ($type eq $CMD_CLIENT_ID) {
	if (! is_client_id(\$value, \$reason)) {
	  print_error_no("$label: $reason");
	  next ;
	}
      } elsif ($type eq $CMD_HOST) {
	if (! is_hostname(\$value, $control, \$reason)) {
	  print_error_no("$label: $reason");
	  next ;
	}
      } elsif ($type eq $CMD_VENDOR) {
	if (! is_vendor(\$value, $control, \$reason)) {
	  print_error_no("$label: $reason");
	  next ;
	}
      } elsif ($type eq $CMD_DOTTED_IP) {
	if (! is_ip(\$value, \$reason)) {
	  print_error_no("$label: $reason");
	  next ;
	}
      } elsif ($type eq $CMD_LIST) {
	if (! is_in_list(\$value, $control, \$reason)) {
	  print_error_no("$label: $reason");
	  next ;
	}
      } else {
	print_error_no("INTERNAL: CONTROL FOR TYPE $type NOT IMPLEMENTED");
	next ;
      }
    }
    
    # update the variable (CMD_ARRAY is processed after the analysis)
    if (($category == $CMD_SIMPLE) || ($category == $CMD_FIELD)) {
      
      $ref = $G_command[$cmdidx]{argument}[$argidx]{variable};
      $$ref = $value;
    } elsif ($category == $CMD_HASH) {
      # it's an hash
      $field=$G_command[$cmdidx]{argument}[$argidx]{field};
      $for_array->{$G_command[$cmdidx]{argument}[$argidx]{field}} = $value;
    }

  }
  
  #
  # check if mandatary arguments are missing or if extra fields are present
  #
  if ($nb_field > $nb_arg) {
    print_error_no("Extra fields are present") ;
  } elsif ($nb_arg > $nb_field) {
    for (; ($argidx < $nb_arg); $argidx++) {
      if ($G_command[$cmdidx]{argument}[$argidx]{mandatory}) {

	$label = build_label($cmdidx, $argidx, $arguments, $nb_field);
	print_error_no("$label is mandatory") ;
      }
    }
  }

  # for an array, push the current element
  if ($category == $CMD_HASH) {
    my $ref = $G_command[$cmdidx]{array};
    push @$ref, $for_array;
  } elsif ($category == $CMD_ARRAY) { 
    my $ref = $G_command[$cmdidx]{array};
    push @$ref, $value;
  }

  # do not call the post control if there is error in the command
  # (fields may be uninitialized)
  if ((! $G_cmd_error) && (defined($G_command[$cmdidx]{post}))) {
    my $rank = @{$G_command[$cmdidx]{array}} - 1;
    &{$G_command[$cmdidx]{post}}($rank);
  }
  
}

#------------------------------------------------------------------------------
#
# analyze the command
#
#------------------------------------------------------------------------------

sub analyze_command {

  my ($arguments) = @_;

  my $i;
  my $found;
  my $cmdidx;

  # check that the command exists
  for ($i = 0, $found = $FALSE; 
       ($i <  (scalar(@G_command)) && (! $found)); 
       $i++) {
    if ($G_command[$i]{name} eq $G_directive) {
      $found = $TRUE;
      $cmdidx = $i;
    }
  }
  
  # do not go forward if the directive is unknown
  if (! $found) {
    print_error_no("$G_directive unknown");
    return;
  }

  # set the flag if the directive has declared one
  if (defined($G_command[$cmdidx]{flag})) {
    my $ref = $G_command[$cmdidx]{flag};
    $$ref = $TRUE;
  }
  # if the syntax is analyze per field
  if ($G_mode == $PER_FIELD) {
    analyze_per_field($cmdidx, $arguments);
  } else {
    print_error_no("PER_KEYWORD not yet implemented");
    return;
  }
       

}

#------------------------------------------------------------------------------
#
# extract the commands from the file and call the command analyzer
#
#------------------------------------------------------------------------------

sub parse_file {

  my ($refcommand, $filename) = @_;

  @G_command = @$refcommand;
  $G_filename = $filename;

  my $line;
  my $word;
  my @arguments = ();
  my @allwords;
  my $directive_found = $FALSE;
  my $continuation_found = $FALSE;

  if (! -r $filename) {
    COMMON::error("$filename is not accessible for reading");
  }

  open(FILE, "<$G_filename") or die("Can't access $G_filename") ;
  $G_lineno = 0;

  while (<FILE>) {
    $G_lineno++;
    chomp($line = $_);

    # remove the comments
    $line =~ s/#.*// ;

    # blanks line are ignored except if continuation of previous line
    # (need to trigger the analysis of the already read command)
    if (($line =~ m/^\s*$/) and (! $continuation_found)) {
      next ;
    }

    $continuation_found = $FALSE;
    # look if the continuation character is at the end (can't be processed
    # correctly by ParseWords so needs to be detected before)
    if ($line =~ m/(.*)(\\[\s\t]*)$/) {
      # the last significant character is the continuation character
      # remove it and set the indicator
      $continuation_found = $TRUE;
      $line = $1;
    }

    if ( !$directive_found) {
      # the line must contain a directive: <name>=<value>
      # otherwise it is a continuation line
      if ($line !~ m/[\s\t]*([^\s\t]*)[\s\t]*$delimiter_string[\s\t]*(.*)/) {
	print_error_no("No directive found");
	next;
      }
      $G_directive=$1;
      $G_directive =~ tr/a-z/A-Z/;
      $directive_found = $TRUE;
      $line = $2;
    }
    
    # all syntaxical elements are extracted to build the command
    # do not remove quotes: sometimes we need to known if it was
    # a quoted string (remove on a case per case basis)
    @allwords = Text::ParseWords::parse_line('\s+', 1, $line);
    my $i;
    for ($i = 0; $i < scalar(@allwords); $i++) {
      if (defined($allwords[$i])) {
	$word = $allwords[$i];
	if (length($word) > 0) {
	  # replace the empty string "" by a null string
	  if ($word eq "\"\"") {
	    $word = "";
	  }
	  push @arguments, $word;
	}
      }
    }
    
    # check for the continuation character at the last position
    if (! $continuation_found) {
      analyze_command(\@arguments);
      @arguments=();
      $directive_found = $FALSE;
    }
    
  }
  
  if ($continuation_found) {
    print_error_no("unexpected end of file");
  }

  return $G_parsing_ok;
}
