#!/usr/bin/perl
# -----------------------------------------------------------------------------
#   ident "@(#)wait_sndr.pl 1.6 04/01/05 SMI"
# -----------------------------------------------------------------------------
#
#  Copyright 2003-2004 Sun Microsystems, Inc. All rights reserved.
#
# -----------------------------------------------------------------------------

package WAIT_SNDR;

use strict;
use FileHandle;
use variables;

require "$ENV{NHINSTALL_LIB}/tools/lib/common.pl";
require "$ENV{NHINSTALL_LIB}/tools/lib/conf.pl";

my $SNDRSTAT="/usr/opt/SUNWesm/sbin/sndrstat -l";

my @SLICE = ();

my $replication_complete = $TRUE;
my $total_size = 0;
my $replicated_size = -1;
my $previous_percentage = -1;

#------------------------------------------------------------------------------
#
# check if a slice already exist otherwise the slice is inserted
# the slice info are then updated
#
#------------------------------------------------------------------------------

sub insert_and_update {

  my ($source, $dest, $size, $percentage, $state) = @_;

  for (my $i = 0 ; $i < scalar(@SLICE) ; $i++) {
    if ($source eq $SLICE[$i]{source}) {
      $SLICE[$i]{percentage} = $percentage;
      $SLICE[$i]{state}      = $state;
      return;
    }
  }

  my $slice = {};
  $slice->{source}     = $source;
  $slice->{dest}       = $dest;
  $slice->{size}       = $size;
  $slice->{percentage} = $percentage;
  $slice->{state}      = $state;
  push @SLICE, $slice;
  $total_size += $slice->{size};

  
}

#------------------------------------------------------------------------------
#
# extract the information regarding the replicated slice
#
#------------------------------------------------------------------------------

sub get_info {

  my ($node_ip) = @_;

  my @result;
  my $i;

  @result = qx/rsh -n $node_ip \"$SNDRSTAT\"/;

  $replicated_size = 0;
  $replication_complete = $TRUE;

  # example of output
  # 
  # both first lines are headers
  #
  #Type    VolGroup                        Vols                            size    s/n  state      dr     dwr     cr     cwr
  #S P    no group         rdsk/c0t0d0s3 => men253-2-cgtp:rdsk/c0t0d0s3    4198392   0.00  REP     2147483647      2147483647      2147483647      2147483647 
  #S P    no group         rdsk/c0t0d0s4 => men253-2-cgtp:rdsk/c0t0d0s4    4198392   0.00  REP     2147483647      2147483647      2147483647      2147483647 

  for ($i = 2; $i < scalar(@result) ; $i++) {
    $result[$i] =~ m/[^\s]*[\s]*[^\s]*[\s]*[^\s]*[\s]*[^\s]*[\s]*([^\s]*)[\s]*[^\s]*[\s]*([^\s]*)[\s]*([^\s]*)[\s]*([^\s]*)[\s]*([^\s]*)/;

    # $1: source
    # $2: dest
    # $3: size
    # $4: percentage
    # $5: state
    insert_and_update($1, $2, $3, $4, $5);

    # compute the total replicated size
    # only for "SYN"
    if ($5 ne "LOG") {
      $replicated_size += $3 * (100 -$4);
    }

    if ($5 ne "REP") {
      # at least one is not yet replicated
      $replication_complete = $FALSE;
    }
  }
}

#------------------------------------------------------------------------------
#
# wait for the SNDR to be finished
#
#------------------------------------------------------------------------------


sub men_wait_sndr {
  
  ($NODE_ID) = @_;

  my $slice;
  my $i;
  
  my $node_ip = COMMON::get_external_ip($NODE_ID);

  if (! $SYNC_FLAG) {
    # sync done on customer's request: do not wait for replication
    COMMON::print_stage("Please execute /opt/SUNWcgha/sbin/nhenablesync to synchronize replicated partitions.");
    return;
  }

  COMMON::print_stage("Waiting for the disks of the master-eligible nodes to synchronize.") ;

  # it may need several seconds before the sndradm command displays
  # at least one useful line
  # we accept 10 seconds without info before displaying that there is
  # a problem
  # otherwise we wait for all the replicated slices to be active
  # with a maximum number of seconds (60)

  COMMON::print_action("Partition detection");

  my $loop = 0;
  my $detected_slice = 0;
  my $expected_slice = 0;
  my $max_no_empty = 10;

  # compute the number of replicated slice
  for $i (0..$#SLICE_LIST) {
    if ($SLICE_LIST[$i]{bitmap} ne "") {
      $expected_slice++;
    }
  }

  my $max_all_detected = 10 * $expected_slice;

  COMMON::print_subaction("$expected_slice partitions expected");


  
  while ($detected_slice < $expected_slice) {
    get_info($node_ip);
    if ($total_size <= 0) {
      if ($loop >= $max_no_empty) {
	COMMON::error("No information about the partitions to be replicated: the Foundation Services probably failed to start on the master-eligible nodes.");
      }
    } else {
      $detected_slice = scalar(@SLICE);
      if ($detected_slice < $expected_slice) {
	if ($loop >= $max_all_detected) {
	  COMMON::error("Some partitions cannot be detected with \"$SNDRSTAT\".");
	}
      }
    }
    
    $loop++;
    sleep 1;
  }

  COMMON::print_action("Partitions to be replicated:");
  # display the slices information
  foreach $slice (@SLICE) {
    # size = number of 512-bytes blocks: / 2 for number of KB / 1000 for the number od MB
    my $size = int($slice->{size} / 2000);
    COMMON::print_subaction("$slice->{source} ($size MB)");
    }
  
  autoflush STDOUT 1;


  for (;;) {
    get_info($node_ip);

    my $replicated_percentage = $replicated_size / $total_size;
    # compute the string to be displayed
    my $begin = int($previous_percentage);
    my $end   = int($replicated_percentage);
    for ($i = $begin+1 ; $i <= $end ; $i++) {
      if ($i % 10 == 0) {
	print "$i%";
      } elsif ($i % 2 == 0) {
	print ".";
      }
    }
    if ($replication_complete) {
      print "\n";
      last ;
    }

    $previous_percentage = $replicated_percentage;
    sleep 1 ;
  }     
  autoflush STDOUT 0;
}

{
}
